import collections
import copy
import os
import traceback
from enum import Enum
from threading import Condition, Thread
from typing import Any, Callable, Dict, Optional, Set, Tuple, Type, List
import logging
from functools import partial
from dataclasses import dataclass
from inspect import signature

import time

from dataspree.platform_sdk.client import Client
from dldstraining.core.exceptions import InterruptedThroughFlag
from dldstraining.core.reference import Ref, as_reference
import weakref

logger = logging.getLogger(__name__)


def once(_function=None, *, check_arguments: bool = True, copy_results: bool = False, deepcopy_results: bool = True,
         as_property: bool = False):
    """
    Cache result if repeatedly called with the same arguments.
    Does not cache results if called with the same arguments in alternating order to reduce space consumption.

    :param _function:           Function if annotated without arguments. Not supplied manually
    :param check_arguments:     If true, return cached results only if the arguments with which the function is
                                currently evaluated don't differ the previous arguments.

    :param copy_results:        True, if you would like the cached results to be soft-copied.

    :param deepcopy_results:    True, if you would like to deepcopy the cached results

    :param as_property:         If true, and the annotated function does not take any arguments, cache the result as
                                a property of this class.

    :return:                    The return value of the annotated function.
    """
    result_copy_function = copy.deepcopy if deepcopy_results else copy.copy if copy_results else lambda a: a

    def _once(function: Callable):
        function_args: List[str] = list(signature(function).parameters.keys())

        def caching_function(*args, **kwargs):
            all_args: Dict[str, Any] = dict(**kwargs, **{function_args[i]: args[i] for i in range(len(args))})

            __cached_arguments_key: str = '_last_arguments'
            __cached_result_key: str = '_return_arguments'
            has_cached_result: bool = hasattr(function, __cached_result_key)
            cached_result: Optional[Any] = getattr(function, __cached_result_key, None)

            if has_cached_result and check_arguments:

                cached_args: Any = getattr(function, __cached_arguments_key)
                assert hasattr(function, __cached_arguments_key) and type(cached_args) == dict

                # Check if the previous arguments match the current arguments
                if (len(cached_args.keys()) == len(all_args)
                    and not any(f not in all_args for f in cached_args.keys())) \
                        and all(v == all_args[k] for k, v in cached_args.items()):
                    return cached_result

            return_value = function(*args, **kwargs)

            setattr(function, __cached_result_key, result_copy_function(return_value))
            setattr(function, __cached_arguments_key, copy.deepcopy(all_args))
            return return_value

        if as_property:
            assert not len(function_args)
            return caching_function()

        return caching_function

    if _function is not None:
        return _once(_function)

    return _once


@dataclass
class HeartbeatState:
    static_args: Dict
    dynamic_args: Optional[Dict] = None
    fun_progress: Optional[Callable] = None
    __old_values = None

    def arguments(self, **additional_arguments):
        evaluated_arguments = self.static_args.copy()
        evaluated_arguments.update({k: v() for k, v in (self.dynamic_args or dict()).items()})
        evaluated_arguments.update(dict(additional_arguments))
        return evaluated_arguments

    def ok(self) -> bool:
        if self.fun_progress is None:
            return True

        old_values = None if self.__old_values is None else dict(self.__old_values)
        new_values = self.__old_values = {k: self.dynamic_args.get(k, partial(self.static_args.get, k))()
                                          for k in signature(self.fun_progress).parameters.keys()}
        return old_values is None or self.fun_progress(**{k: (old_values.get(k), v) for k, v in new_values.items()})

    def validate(self) -> 'HeartbeatState':
        assert self.dynamic_args is None or all(map(callable, self.dynamic_args.values())), \
            f'Some dynamic arguments are not references: ' \
            f'{list(zip(self.dynamic_args.keys(), map(callable, self.dynamic_args.values())))}.'

        # Check that there are no missing arguments for the progress function that is to be evaluated
        if __debug__:
            if self.fun_progress is not None:
                missing_args = [f'"{k}"' for k in signature(self.fun_progress).parameters.keys()
                                if k not in self.dynamic_args.keys() and k not in self.static_args.keys()]
                assert not len(missing_args), f'Missing validation argument{"s: " * (min(len(missing_args), 1) - 1)} ' \
                                              f'{", ".join(missing_args)}.'
        return self

    def copy(self) -> 'HeartbeatState':
        hbs = copy.copy(self)
        assert hbs.__old_values == self.__old_values

        hbs.static_args = copy.deepcopy(self.static_args)
        hbs.dynamic_args = dict(self.dynamic_args or {})

        return hbs

    def __str__(self) -> str:
        return ', '.join((f'{k}={value}' for k, value in (self.arguments()).items()))


class HeartbeatTrace:

    def __init__(self, heartbeat_context, trace: bool = True, *_args, **_kwargs) -> None:
        self.__heartbeat_context = heartbeat_context
        self.__trace: bool = trace

    def __evaluate_function(self, function_name: str, member_function: Callable, *args, **kwargs) -> Any:
        try:
            self.__heartbeat_context.update_trace(function_name, False)
            return member_function(*args, **kwargs)
        finally:
            self.__heartbeat_context.update_trace(function_name, True)

    def __getattribute__(self, name: str) -> Any:
        attr = super().__getattribute__(name)

        if name.startswith('_'):
            return attr

        if not self.__trace or not hasattr(attr, '__call__'):
            return attr

        return partial(self.__evaluate_function, name, attr)


class HeartbeatContext:
    """
    The HeartbeatContext can be used to execute the provided heartbeat_function at the provided heartbeat_interval in a
    separate thread.

    It is recommended to use the method wrap_with_heartbeat in a context manager context:
    >>> heartbeat_function: Callable[]
    >>> heartbeat_interval: float
    >>> with HeartbeatContext.automatic(heartbeat_function, heartbeat_interval):
    >>>     # do something
    >>>     pass
    """

    class Mode(Enum):
        MANUAL = 'manual   '
        WATCH = 'watch    '
        AUTOMATIC = 'automatic'

    def __init__(self, heartbeat_function: Callable, heartbeat_interval: float, task: str, *,
                 running_condition: Optional[Callable[[], bool]] = None,
                 exceptions_to_report: Optional[Set[Type[Exception]]] = None):
        """
        Create a heartbeat context.
        This context becomes active, once it is used inside of a context manager, or if the activation methods are
        called.

        :param heartbeat_function:      The function that is to be executed whenever a heartbeat occurs.

        :param heartbeat_interval:      Duration from one automatic heartbeat to another.

        :param task:                    String that represents the task that is currently executed.

        :param running_condition:       If set, this function is evaluated to determine if the heartbeat
                                        loop has to stop without exiting the context and without and exception being
                                        thrown.


        :param exceptions_to_report:    Set of exceptions that are reported in case this instance is used as a
                                        context manager and the exception escapes the scope of the contextManager.

        """
        self.heartbeat_function: Callable = heartbeat_function
        self.heartbeat_interval: float = heartbeat_interval
        self.task: str = task
        self.exceptions_to_report: Set[Type[Exception]] = exceptions_to_report or {BaseException}

        self.running_condition: Optional[Callable[[], bool]] = running_condition

        self.running: bool = False
        self.thread: Optional[Thread] = None

        # (total number of operations executed since start, name of the operation, start time of the operation)
        self.current_operation: Tuple[int, str, float] = (0, '/', time.time())

        # (name of the function that is executed,
        #  execution count of this particular function,
        #  call completed,
        #  start / stop time of the operation)
        self.current_trace: Optional[Tuple[str, int, bool, float]] = None

        # name -> exited, specific function execution count, elapsed_time total, last trigger time
        self.__traces: Dict[str, Tuple[bool, int, float, float]] = dict()

        self.cv = Condition()
        self.heartbeat_state: Tuple[Optional[HeartbeatContext.Mode], HeartbeatState, Optional[float]] = (
            None, HeartbeatState(dict()), None)
        self.heartbeat_states = collections.deque()
        self.__time_last_beat: float = 0

    @staticmethod
    @as_reference({'references'})
    def manual(heartbeat_function: Callable, heartbeat_interval: float, task: str,
               status: str, running_condition: Optional[Callable[[], bool]] = None, operation: str = '', *,
               exceptions_to_report: Optional[Set[Type[Exception]]] = None,
               references: Optional[Dict[str, Ref]] = None, **kwargs) -> 'HeartbeatContext':
        """
        Manually call this function to generate manual heartbeats by updating the operation.

        :param operation:               Name of the operation that is performed.

        :param heartbeat_function:      The function that is to be executed whenever a heartbeat occurs.

        :param heartbeat_interval:      Duration from one automatic heartbeat to another.

        :param running_condition:       If set, this function is evaluated to determine if the heartbeat
                                        loop has to stop without exiting the context and without and exception being
                                        thrown.

        :param exceptions_to_report:    Set of exceptions that are reported in case this instance is used as a
                                        context manager and the exception escapes the scope of the contextManager.

        :param task:                    String that represents the (long-running) task that is currently executed (
                                        e.g., training).

        :param status:                  String that represents the status (e.g., running / terminated)

        :param operation:               The current operation (e.g., initialize trainer)

        :param references:              Values that are to be supplied to the heartbeat function as references.

        :param kwargs:                  Values that are to be supplied to the heartbeat function as references.

        :return:                        The instance of the HeartbeatLogger.
        """
        hbc = HeartbeatContext(heartbeat_function=heartbeat_function, heartbeat_interval=heartbeat_interval, task=task,
                               running_condition=running_condition, exceptions_to_report=exceptions_to_report)
        return hbc.activate_manual_mode(operation=operation, references=references, status=status, **kwargs)

    @as_reference({'references'})
    def activate_manual_mode(self, operation: str, references: Optional[Dict[str, Ref]] = None, *,
                             inherit: bool = True, **kwargs) -> 'HeartbeatContext':
        """ @see manual """
        self.__update_state(new_mode=HeartbeatContext.Mode.MANUAL,
                            new_static_args=dict(kwargs), new_dynamic_args=references, fun_progress=None,
                            inherit=inherit)
        return self.update_operation(operation, inform=False)

    @staticmethod
    @as_reference({'references'})
    def automatic(heartbeat_function: Callable, heartbeat_interval: float, task: str, status: str,
                  running_condition: Optional[Callable[[], bool]] = None, operation: str = '', *,
                  exceptions_to_report: Optional[Set[Type[Exception]]] = None,
                  references: Optional[Dict[str, Ref]] = None, **kwargs) -> 'HeartbeatContext':
        """
        Automatically send a heartbeat, until this mode is terminated by an exception that escapes the ContextManger,
        a False running_condition, or by manually selecting a different HeartbeatMode.

        Try to avoid this mode whenever possible, because it automatically sends out heartbeats.
        The HeartbeatContext can only detect that something bad happened if an exception is thrown that would escape
        the ContextManager in which the HeartbeatContext is used.

        :param operation:               Name of the operation that is performed.

        :param heartbeat_function:      The function that is to be executed whenever a heartbeat occurs.

        :param heartbeat_interval:      Duration from one automatic heartbeat to another.

        :param running_condition:       If set, this function is evaluated to determine if the heartbeat
                                        loop has to stop without exiting the context and without and exception being
                                        thrown.

        :param exceptions_to_report:    Set of exceptions that are reported in case this instance is used as a
                                        context manager and the exception escapes the scope of the contextManager.

        :param task:                    String that represents the (long-running) task that is currently executed (
                                        e.g., training).

        :param status:                  String that represents the status (e.g., running / terminated)

        :param operation:               The current operation (e.g., initialize trainer)

        :param references:              Values that are to be supplied to the heartbeat function as references.

        :param kwargs:                  Values that are to be supplied to the heartbeat function as references.

        :return:                        The instance of the HeartbeatLogger.
        """
        hbc = HeartbeatContext(heartbeat_function=heartbeat_function, heartbeat_interval=heartbeat_interval, task=task,
                               running_condition=running_condition, exceptions_to_report=exceptions_to_report)
        return hbc.activate_automatic_heartbeat_mode(operation=operation, references=references, status=status,
                                                     **kwargs)

    @as_reference({'references'})
    def activate_automatic_heartbeat_mode(self, operation: str, references: Optional[Dict[str, Ref]] = None, *,
                                          inherit: bool = True, timeout: Optional[float], **kwargs) \
            -> 'HeartbeatContext':
        """
        @see automatic
        """
        self.__update_state(new_mode=HeartbeatContext.Mode.AUTOMATIC,
                            new_static_args=dict(kwargs), new_dynamic_args=references, fun_progress=None,
                            inherit=inherit, max_duration=timeout)

        return self.update_operation(operation, inform=False)

    @staticmethod
    @as_reference({'references'})
    def watch(heartbeat_function: Callable, heartbeat_interval: float, task: str, status: str,
              evaluate_progress: Callable[[Any], bool], running_condition: Optional[Callable[[], bool]] = None,
              operation: str = '', *, exceptions_to_report: Optional[Set[Type[Exception]]] = None,
              references: Optional[Dict[str, Ref]] = None, **kwargs) -> 'HeartbeatContext':
        """
        Send a heartbeat message whenever a the function "evaluate_progress" determines that progress was made.

        :param evaluate_progress:       Function that is used to evaluate the progress.
                                        The function has to take as a named arguments values from the set of
                                        references.
                                        For each specified argument, the heartbeat context passes a tuple that
                                        consists of the previous assignment and the current assignment, that together
                                        allow the function to determine if progress was made.

                                        Function: {keyword_1=(previous_value, current_value), keyword_2 = ...} -> bool

                                        Example assignment:
                                        lambda iteration: iteration[0] is None  or iteration[1] > iteration[0]

                                        if the key "iteration" is in the set of references.

        :param operation:               Name of the operation that is performed.

        :param heartbeat_function:      The function that is to be executed whenever a heartbeat occurs.

        :param heartbeat_interval:      Duration from one automatic heartbeat to another.

        :param running_condition:       If set, this function is evaluated to determine if the heartbeat
                                        loop has to stop without exiting the context and without and exception being
                                        thrown.

        :param exceptions_to_report:    Set of exceptions that are reported in case this instance is used as a
                                        context manager and the exception escapes the scope of the contextManager.

        :param task:                    String that represents the (long-running) task that is currently executed (
                                        e.g., training).

        :param status:                  String that represents the status (e.g., running / terminated)

        :param operation:               The current operation (e.g., initialize trainer)

        :param references:              Values that are to be supplied to the heartbeat function as references.

        :param kwargs:                  Values that are to be supplied to the heartbeat function as references.

        :return:                        The instance of the HeartbeatLogger.
        """
        hbc = HeartbeatContext(heartbeat_function=heartbeat_function, heartbeat_interval=heartbeat_interval, task=task,
                               running_condition=running_condition, exceptions_to_report=exceptions_to_report)
        return hbc.activate_watch_mode(operation=operation, evaluate_progress=evaluate_progress,
                                       references=references, status=status, **kwargs)

    @as_reference({'references'})
    def activate_watch_mode(self, operation: str, evaluate_progress: Callable[[Dict], bool], references: Dict[str, Ref],
                            *, inherit: bool = True, **kwargs) -> 'HeartbeatContext':
        """
        @see watch
        """
        self.__update_state(new_mode=HeartbeatContext.Mode.WATCH, new_static_args=dict(**kwargs),
                            new_dynamic_args=references, fun_progress=evaluate_progress, inherit=inherit)

        return self.update_operation(operation, inform=False)

    def update(self, *, operation: Optional[str] = None, references: Optional[Dict[str, Ref]] = None,
               **static_args: Any) -> None:
        """
        Update parts of the state

        :param operation:
        :param references:
        :param static_args:
        :return:
        """
        self.__update_state(new_dynamic_args=references, new_static_args=dict(static_args))
        self.update_operation(operation)

    def update_operation(self, new_operation: Optional[str], inform: bool = True) -> 'HeartbeatContext':
        """
        Update the currently active operation.

        :param new_operation:   The new operation string.
        """
        if new_operation is not None:
            self.current_operation = self.current_operation[0] + 1, new_operation, time.time()
            if inform:
                logger.info(f'Active operation: {new_operation}.')

        # Report the information to the loop (if running) and the logger.
        self.__commit_change(operation=new_operation is not None)

        return self

    def update_trace(self, trace_name, exited: bool = False):
        # Update the entry in the trace lookup table of this HeartbeatContext instance.
        if trace_name not in self.__traces:
            assert not exited
            self.__traces[trace_name] = (True, 0, 0.0, time.time())

        _exited, count, elapsed_time, last_time = self.__traces[trace_name]

        if not exited:
            count += 1
        else:
            elapsed_time += (time.time() - last_time)

        self.__traces[trace_name] = (exited, count, elapsed_time, time.time())

        # Update the current trace.
        self.current_trace = (trace_name, count, exited, time.time())

        # Report the information to the loop (if running) and the logger.
        self.__commit_change(operation=False)

    def __commit_change(self, operation: bool = False) -> None:

        # If we are not in manual mode, update the heartbeat state.
        # Else, send a heartbeat because either the operation or the trace have updated.
        currently_active_mode = self.heartbeat_state[0] if self.heartbeat_state is not None else None
        if currently_active_mode == HeartbeatContext.Mode.MANUAL:

            if operation:
                self.__beat(self.heartbeat_state, log_all_info=True)

        else:
            sent_trace_info: Dict[str, str] = dict() if self.current_trace is None else dict(
                trace=self.current_trace[0])

            self.__update_state(**sent_trace_info)
            self.__log_heartbeat(self.heartbeat_state, **sent_trace_info)

        if operation:
            # Notify the loop
            with self.cv:
                self.cv.notify_all()

    def __update_state(self, *, new_mode: Optional[Mode] = None, new_dynamic_args: Optional[Dict[str, Ref]] = None,
                       new_static_args: Optional[Dict[str, Any]] = None, inherit: bool = True,
                       max_duration: Optional[float] = None, **kwargs: Any) -> None:

        # Update Mode
        if new_mode is not None:
            self.heartbeat_state = new_mode, self.heartbeat_state[1], None if max_duration is None \
                else time.time() + max_duration

        # Update State dynamic and static args
        if self.heartbeat_state[1].dynamic_args is None or not inherit:
            self.heartbeat_state[1].dynamic_args = dict()
        self.heartbeat_state[1].dynamic_args.update(new_dynamic_args if new_dynamic_args is not None else dict())

        if self.heartbeat_state[1].static_args is None or not inherit:
            self.heartbeat_state[1].static_args = dict()
        self.heartbeat_state[1].static_args.update(new_static_args if new_static_args is not None else dict())

        # Update progress if to be set
        if 'fun_progress' in kwargs:
            self.heartbeat_state[1].fun_progress = kwargs['fun_progress']
            self.heartbeat_state[1].validate()

    def __run(self) -> None:
        """
        Run in case we have got an automatic heartbeat (or a heartbeat that depends on a condition).
        Wait otherwise.

        :return:
        """
        self.running = True
        try:
            while self.running:
                with self.cv:

                    if not self.running:
                        continue

                    # Manual mode.
                    hbs = self.heartbeat_state
                    if hbs is None or hbs[0] == HeartbeatContext.Mode.MANUAL:
                        self.cv.wait()
                        continue

                    hbs = self.heartbeat_state
                    if hbs is None:
                        continue

                    mode, state, timeout = hbs
                    state = state.copy()

                self.__beat((mode, state, timeout))

                time.sleep(self.heartbeat_interval)

        except InterruptedThroughFlag:
            pass

        finally:
            self.running = False

    def __beat(self, heartbeat_state: Tuple[Optional[Mode], HeartbeatState, Optional[float]], *,
               log_all_info: bool = False, **kwargs) \
            -> bool:

        operation_count, operation_name, _operation_time_start = self.current_operation

        if log_all_info:
            self.__log_heartbeat(heartbeat_state, **kwargs)

        currently_active_mode, state, timeout = heartbeat_state

        if timeout is not None and timeout < time.time():
            raise InterruptedThroughFlag(f'Operation {currently_active_mode.value} timed out.')

        if currently_active_mode is None:
            currently_active_mode_str = 'error'
        else:
            currently_active_mode_str = currently_active_mode.value

        ok = state.ok()
        if ok:
            information: Dict[str, Any] = state.arguments(mode=currently_active_mode_str.strip(),
                                                          task=self.task, operation=operation_name, **kwargs)

            self.heartbeat_function(information)

            logger.debug(''.join(f'{k}={value}' for k, value in information.items()))
            information['time_since_last_beat'] = time.time() - self.__time_last_beat
            self.__time_last_beat = time.time()

            if __debug__:
                with open(self.__log_file_send(), 'a') as _log_file:
                    _log_file.write(f'{information}\n')

        if self.__demands_interruption():
            raise InterruptedThroughFlag()

        return ok

    def __log_heartbeat(self, heartbeat_state: Tuple[Optional[Mode], HeartbeatState, Optional[float]],
                        **kwargs) -> None:
        # Log information
        operation_count, operation_name, _operation_time_start = self.current_operation

        currently_active_mode, state, timeout = heartbeat_state
        currently_active_mode_value = currently_active_mode.value if currently_active_mode is not None else 'None'
        information: Dict[str, Any] = state.arguments(mode=currently_active_mode_value.strip(), **kwargs)

        # Remove redundant information
        for key in ('trace', 'operation', 'mode', 'task'):
            information.pop(key, None)

        if timeout is not None:
            information['time until timeout'] = timeout - time.time()

        trace_info: str = ''
        sent_trace_info: Dict[str, str] = dict()
        if self.current_trace is not None:
            trace_name, specific_trace_count, trace_exited, _trace_time_triggered = self.current_trace
            trace_exited_str = '[terminated]' if trace_exited else '[started]   '
            avg_trace_time = self.__traces[trace_name][2] / self.__traces[trace_name][1]

            trace_info = f'{trace_exited_str} {trace_name} ({specific_trace_count:3}), ' \
                         f'avg. proc. time {int(avg_trace_time * 1000):4}ms.'

            sent_trace_info['trace'] = trace_name

        txt: str = f'{currently_active_mode_value} ' \
                   f'Task {self.task} ' \
                   f'Operation {operation_count} ({operation_name}) {trace_info} {state} '
        txt += ', '.join(f'{k}={value}' for k, value in information.items())

        if __debug__:
            with open(self.__log_file(), 'a') as _log_file:
                _log_file.write(f'{txt}\n')

    def __enter__(self) -> 'HeartbeatContext':
        """


        :return:        A weak reference to this object. This way, the object cannot be used anymore when outside
                        the scope of the context manager.
        """
        self.heartbeat_states.append(self.heartbeat_state)

        if not self.running:

            if self.thread is not None and self.thread.is_alive():
                self.thread.join(timeout=2)
                if self.thread.is_alive():
                    raise TimeoutError('The heartbeat loop was unable to join the heartbeat thread.')

            self.thread = Thread(target=self.__run)
            self.thread.start()

        return weakref.proxy(self)

    def __exit__(self, ex_type, exc_value, tb) -> Optional[bool]:
        """
        Called whenever the context of this heartbeat context exits.

        :param ex_type:
        :param exc_value:
        :param tb:
        :return:
        """

        self.heartbeat_state = self.heartbeat_states.pop()

        tb_stack: List[str] = None if tb is None else traceback.format_tb(tb)

        if ex_type is not None and any(filter(partial(issubclass, ex_type), self.exceptions_to_report)):
            try:
                self.__beat(self.heartbeat_state, log_all_info=True, status='running',
                            error=str(exc_value), error_type=getattr(ex_type, '__name__', ''), traceback=tb_stack)
            except InterruptedThroughFlag:
                pass

        terminates_loop: bool = ex_type is not None or not len(self.heartbeat_states)

        if terminates_loop:
            self.running = False
            self.heartbeat_state = None, HeartbeatState(dict()), None
            with self.cv:
                self.cv.notify_all()

            if self.thread.is_alive():
                self.thread.join(timeout=5.0)
                if self.thread.is_alive():
                    raise TimeoutError('The heartbeat loop was unable to join the heartbeat thread after 5 seconds.')

            # Catch the exception "InterruptedThroughFlag" if this only in the last context manager that is active.
            if ex_type == InterruptedThroughFlag:
                return True

        else:

            # Update the currently active mode and heartbeat state
            self.heartbeat_state = self.heartbeat_states[-1]

            with self.cv:
                self.cv.notify_all()

        if not len(self.heartbeat_states):
            # TODO: do we report "finished" in case of an Exception?
            try:

                self.__beat(self.heartbeat_state, log_all_info=True,
                            status='finished' if ex_type is None else 'exception')
            except InterruptedThroughFlag:
                pass

        # We don't really need this, if the error is important we will be notified either way.
        # if tb_stack is not None and ex_type != InterruptedThroughFlag:
        #    logger.error('An exception occurred: ' + ''.join(tb_stack))

    def __demands_interruption(self) -> bool:
        return self.running_condition is not None and not self.running_condition()

    @staticmethod
    @once
    def __log_file() -> str:
        file_name: str = os.path.join(Client.dataspree_temp_dir('heartbeat'), 'log.txt')
        with open(file_name, 'w'):
            return file_name

    @staticmethod
    @once
    def __log_file_send() -> str:
        file_name: str = os.path.join(Client.dataspree_temp_dir('heartbeat'), 'send_log.txt')
        with open(file_name, 'w'):
            return file_name
