import copy
import itertools
import json
import logging
import os
import time
import math
from typing import Any, Dict, List, Optional, Tuple
from zipfile import ZipFile

import onnx
from pytorch_lightning import Callback, seed_everything

from ..core.reference import Ref
from ..models.flexnet import onnx_export_flexmodel

from dldstraining.core.heartbeat_loop import HeartbeatContext, HeartbeatTrace
from dldstraining.core.utils import AsyncDispatch

logger = logging.getLogger(__name__)


class DLDSCallback(Callback, HeartbeatTrace):

    def __init__(self, trainer_model, heartbeat_context: HeartbeatContext) -> None:
        super().__init__(heartbeat_context)
        self.trainer_model = trainer_model
        self.heartbeat_context: HeartbeatContext = heartbeat_context

        self.onnx_opset_version: int = 9

        self.upload_checkpoint_element = None

        self.upload_model_element = None

        self.last_checkpoint_time: float = time.time() - 1000
        self.last_log_time: float = time.time() - 1000

        #if trainer.is_global_zero: #XXX Muss das nicht das hier sein? oder gibt es da einen unterschied?
        #if trainer.is_global_zero: #TODO:
        # Das muss passen zu dem join vorgang, aber ist noch nicht initializiert
        # if self.trainer_model.trainer.is_global_zero:


    def on_init_start(self, trainer):
        pass

    def on_init_end(self, trainer):
        pass

    def on_keyboard_interrupt(self, trainer, pl_module):
        pass

    def on_fit_start(self, trainer, pl_module):
        pass

    def start_waiting_for_dataset_items(self, operation: str) -> None:
        """
        We call this callback whenever we might wait for dataset items to be loaded.
        """
        self.heartbeat_context.activate_automatic_heartbeat_mode(operation=f'{operation} load items', timeout=60 * 10)


    def on_train_epoch_start(self, trainer, pl_module):
        # Called before downloading dataset items for the training
        self.start_waiting_for_dataset_items('train')


    def on_train_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        # adjust learning rate according to schedule
        pl_module.lr_scheduler.step(iteration=trainer.global_step)

    def on_train_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
        if pl_module.hparams.model.get('ema', False):
            tau: float = pl_module.hparams.model.get('tau', 0.9998)
            for param, target_param in zip(pl_module.model.parameters(), pl_module.ema_model.parameters()):
                target_param.data.copy_(tau * target_param.data + (1 - tau) * param.data)

        if trainer.is_global_zero:

            # Log training progress to stdout
            if time.time() - self.last_log_time > self.trainer_model.log_interval_s:
                logger.info(self.trainer_model.create_progress_string())
                self.last_log_time = time.time()

            # If this was the last training batch in current evaluation interval:
            if trainer.global_step != 0 and trainer.global_step % pl_module.hparams.evaluation['interval'] == 0:
                if time.time() - self.last_checkpoint_time >= self.trainer_model.checkpoint_interval_s:
                    self.__evaluate_and_upload(trainer, pl_module)
                    self.last_checkpoint_time = time.time()

    def on_train_start(self, trainer, pl_module):
        if trainer.is_global_zero:
            self.heartbeat_context.activate_watch_mode(operation='train', references=dict(),
                                                       evaluate_progress=lambda iteration:
                                                       iteration[0] is None or iteration[1] > iteration[0])
        if self.upload_checkpoint_element is None or self.upload_checkpoint_element.is_joined():
            self.upload_checkpoint_element: AsyncDispatch[Tuple[int, int, str]] = AsyncDispatch(
                self.upload_checkpoint_from_file, Ref(self.trainer_model, 'running'))
        else:
            logger.warning('upload checkpoint element is neither None nor joined. Implementation logic error.')

        if self.upload_model_element is None or self.upload_model_element.is_joined():
            self.upload_model_element: AsyncDispatch[Tuple[int, Any, str, Dict, Dict, Dict]] = AsyncDispatch(
                self.upload_model_from_file, Ref(self.trainer_model, 'running'))
        else:
            logger.warning('upload model element is neither None nor joined. Implementation logic error.')

        if os.environ.get('NODE_RANK', '0') == '0':
            self.upload_checkpoint_element.start()
            self.upload_model_element.start()
        seed_everything(pl_module.hparams.training.get('seed', 42))
        self.trainer_model.train_start_time = time.time()

    def on_train_end(self, trainer, pl_module):
        logger.info(f'Finishing the training run...')
        if trainer.is_global_zero and self.trainer_model.running:

            if trainer.global_step >= trainer.max_steps - pl_module.trainer.accumulate_grad_batches:
                logger.info('Final evaluation ...')
                self.__evaluate_and_upload(trainer, pl_module)

        if trainer.is_global_zero:
            with self.heartbeat_context.activate_automatic_heartbeat_mode(operation='post train (waiting for upload)',
                                                                          timeout=None):
                logger.info('Waiting for checkpoint and model uploads to finish...')
                if self.upload_checkpoint_element is not None:
                    self.upload_checkpoint_element.join_after_last()
                if self.upload_model_element is not None:
                    self.upload_model_element.join_after_last()
                logger.info('Uploads finished.')

        #self.trainer_model.running = False
        logger.info('Uploads finished. Shutting down...')

        # Clear temporary files.
        for item in itertools.chain(os.listdir(self.trainer_model.checkpoint_dir),
                                    (self.trainer_model.checkpoint_path, self.trainer_model.checkpoint_meta_path)):
            try:
                os.remove(os.path.join(self.trainer_model.checkpoint_dir, item))
            except OSError:
                pass

    def on_validation_start(self, trainer, pl_module):
        # This leads to an evaluation in the same order.
        seed_everything(pl_module.hparams.training.get('seed', 42))
        logger.info(f'Starting the evaluation...')

    def on_validation_batch_start(self, trainer, pl_module, batch, batch_idx, dataloader_idx):
        total_number_of_batches: int = math.ceil(len(pl_module.dataset.test_indices) / pl_module.batch_size)

        if trainer.is_global_zero:
            self.heartbeat_context.activate_watch_mode(
                operation='eval', validation_iteration=batch_idx, validation_end_iteration = total_number_of_batches,
                validation_progress=f'{self.trainer_model.validation_progress(batch_idx, total_number_of_batches): .2f}',
                evaluate_progress=lambda validation_iteration:
                validation_iteration[0] is None or validation_iteration[1] > validation_iteration[0],
                references=dict())

        # Log validation progress to stdout
        if time.time() - self.last_log_time > self.trainer_model.log_interval_s:
            logger.info(self.trainer_model.create_validation_progress_string(batch_idx, total_number_of_batches))
            self.last_log_time = time.time()

    def on_validation_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
        self.start_waiting_for_dataset_items('eval')

    def on_validation_end(self, trainer, pl_module):
        if trainer.is_global_zero:
            self.heartbeat_context.activate_watch_mode(operation='train', references=dict(),
                                                       evaluate_progress=lambda iteration:
                                                       iteration[0] is None or iteration[1] > iteration[0])

    def __evaluate_and_upload(self, trainer, pl_module):
        logger.info('Evaluating model...')
        trainer.run_evaluation()

        model_id = pl_module.platform_model['id']
        iteration = trainer.global_step
        logger.info(f'Triggering upload of models and checkpoint {model_id} in iteration {iteration}...')
        # Save checkpoint to be uploaded
        checkpoint_save_path = os.path.join(self.trainer_model.checkpoint_dir, f'model{iteration}.ckpt')
        trainer.save_checkpoint(checkpoint_save_path)

        if self.upload_checkpoint_element is None or self.upload_checkpoint_element.is_joined():
            logger.error('Cannot upload checkpoint. Implementation error.')
        else:
            self.upload_checkpoint_element.update((model_id, iteration, checkpoint_save_path))

        # Create content of class_label.json
        class_labels: Dict[str, List] = {str(cid): [int(dlds_id), pl_module.dataset.class_name_map[dlds_id]]
                                         for cid, dlds_id in pl_module.dataset.inverse_class_id_map.items()}

        # Create content of parameters.json
        parameters = copy.deepcopy(pl_module.platform_model)
        for key in ('created_by', 'pretrained_checkpoint_id', 'progress', 'inference_data', 'updated_date', 'status',
                    'model_data', 'created_date'):
            parameters.pop(key, None)

        # Create content of model.onnx
        onnx_save_path: str = os.path.join(self.trainer_model.checkpoint_dir, 'model.onnx')
        onnx_export_flexmodel(pl_module.ema_model if pl_module.hparams.model.get('ema', False) else pl_module.model,
                              name=onnx_save_path, verbose=False, opset_version=self.onnx_opset_version)

        # Add metadata to model and parameters
        onnx_model, parameters = self.__add_metadata(self.trainer_model.platform_model, pl_module, iteration,
                                                     parameters=parameters, onnx_path=onnx_save_path)

        if self.upload_model_element is None or self.upload_model_element.is_joined():
            logger.error('Cannot upload model element. Implementation error.')
        else:
            self.upload_model_element.update((model_id, onnx_model, onnx_save_path, parameters,
                                            class_labels, pl_module.preprocessing_augmentations_replay))
        logger.info('Triggered uploading models and checkpoint.')

    @staticmethod
    def __add_metadata(platform_model, pl_module, iteration: int, *,
                       parameters: Optional[Dict] = None, onnx_path: Optional[str] = None) -> Tuple[Any, Dict]:
        onnx_model = onnx.load(onnx_path) if onnx_path is not None else None

        def add(key, value, model: bool = True, param: bool = True):
            if onnx_model is not None and model:
                meta = onnx_model.metadata_props.add()
                meta.key = key
                meta.value = value
            if parameters is not None and param:
                parameters[key] = value

        add('ai_platform_model_id', str(platform_model.get('id', 0)), param=False)
        add('iteration', str(iteration))
        add('network_type', str(platform_model.get('network_type')), param=False)
        add('network_config_option_name', str(platform_model.get('network_config_option_name')), param=False)
        add('input_image_format', list(pl_module.hparams.dataset.get('format', {'RGB': True}).keys())[0], param=False)

        if 'detection' in pl_module.hparams:
            add('scale', str(pl_module.scale), param=False)

            # Output mappings
            maps = ['width', 'height']
            if pl_module.rotation_regression:
                if pl_module.hparams.detection.get('oriented_boxes').get('total_orientation', False):
                    maps.extend(['sin(rotation)', 'cos(rotation)'])
                else:
                    maps.extend(['sin(2*rotation)', 'cos(2*rotation)'])

            add('output_mapping', json.dumps({
                pl_module.model.output_list[0]: [f'class_{i}' for i in range(pl_module.detection_classes)],
                pl_module.model.output_list[1]: maps,
                pl_module.model.output_list[2]: ['offset_x', 'offset_y']}))

        if 'classification' in pl_module.hparams:
            # classification layer name
            add('classification_layer', str(pl_module.model.base.fc_layer))

            # Output mappings
            add('output_mapping', json.dumps({
                pl_module.model.output_list[0]: [f'class_{i}' for i in range(pl_module.classification_classes)]}))

        return onnx_model, parameters

    def upload_checkpoint_from_file(self, model_id: int, iteration: int, current_path: str) -> bool:
        # Move the checkpoint from its previous location to its next location.
        upload_file_path = current_path.replace(f'model{iteration}', f'model{model_id}')

        try:
            os.remove(upload_file_path)
        except OSError:
            pass

        try:
            os.replace(current_path, upload_file_path)
        except OSError as e:
            logger.warning(f'Could not replace {current_path} to {upload_file_path}: {e}.'
                           f'Reattempt if no newer checkpoint is available.')

            return False

        # Upload the model checkpoint from the new location
        result = None
        rex = None
        try:
            result = self.trainer_model.client.upload_latest_model_checkpoint(model_id, iteration, upload_file_path)
        except Exception as e:
            rex = e

        result = result is not None
        if result:
            logger.info(f'Checkpoint {iteration} uploaded.')
        else:
            logger.warning(f'Uploading checkpoint {iteration} failed: {rex}.')

        # Remove the file from the new location
        try:
            if not result:
                try:
                    os.replace(upload_file_path, current_path)
                except OSError:
                    # If this operation results in an error don't reattempt because we lost the file.
                    result = True

            os.remove(upload_file_path)
        except OSError:
            pass

        return result

    def upload_model_from_file(self, model_id: int, onnx_model, onnx_save_path: str, parameters: Dict,
                               class_labels: Dict, preprocessing_augmentations_replay: Dict) -> bool:
        # Write model
        if onnx_model is not None:
            onnx.save_model(onnx_model, onnx_save_path)

        # Create zip file
        logger.info('Model upload: Creating Zip file...')
        upload_file_path: str = os.path.join(self.trainer_model.checkpoint_dir, f'model{model_id}.zip')
        with ZipFile(upload_file_path, 'w') as zf:
            zf.write(onnx_save_path, 'model.onnx')
            zf.writestr('parameters.json', json.dumps(parameters, indent=2))
            zf.writestr('class_label.json', json.dumps(class_labels, indent=2))
            zf.writestr('preprocessing.json', json.dumps(preprocessing_augmentations_replay, indent=2))
        logger.info('Model upload: Uploading  Zip file...')

        # Upload the model checkpoint from the new location
        result = None
        rex = None
        try:
            result = self.trainer_model.client.upload_inference_data(model_id, upload_file_path)
        except Exception as e:
            rex = e
        if result:
            logger.info('Model upload: Zip file uploaded.')
        else:
            logger.warning(f'Model upload: Zip file upload failed: {rex}.')

        # Remove the file from the new location
        try:
            os.remove(upload_file_path)
        except OSError:
            pass

        return result


